@@ -218,7 +218,6 @@ def select_features(X, y, metric, metric_name = '', N = 30, feature_names = None
 
     return idx
 
-
 def BER(X ,y, nobs = 10000, NSigma = 10, show = False, save_fig = ''):
     """
     We draw random samples from the bayes distribution models to calculate BER
@@ -249,7 +248,7 @@ def BER(X ,y, nobs = 10000, NSigma = 10, show = False, save_fig = ''):
     
     #print(mu1, mu2)
     #print(s1, s2)
-    #print(nb.sigma_)
+    #print(nb.var_)
     
     lb = np.minimum(mu1 - NSigma*s1, mu2 - NSigma*s2)
     ub = np.maximum(mu1 + NSigma*s1, mu2 + NSigma*s2)
@@ -277,10 +276,10 @@ def BER(X ,y, nobs = 10000, NSigma = 10, show = False, save_fig = ''):
         
     if X.shape[1] == 2:
         # for 2-dimensional data, plot the contours        
-        ax = plot_gaussian_contour (X, y, nb.theta_[0], nb.sigma_[0], nb.theta_[1], nb.sigma_[1], alpha = 0.3)
+        ax = plot_gaussian_contour (X, y, nb.theta_[0], np.sqrt(nb.var_[0]), nb.theta_[1], np.sqrt(nb.var_[1]), alpha = 0.3)
         plotComponents2D(X, y, set(y), use_markers = False, ax = ax)
         plt.legend()
-        title = ' $ \mu $ = ' + str(np.round(nb.theta_, 3) ) + ', $\sigma$ = ' + str( np.round(nb.sigma_,3) ).replace('\n','')
+        title = ' $ \mu $ = ' + str(np.round(nb.theta_, 3) ) + ', $\sigma^2$ = ' + str( np.round(nb.var_,3) ).replace('\n','')
         plt.title(title)
         
         if (save_fig != '' and save_fig != None):   
@@ -522,7 +521,7 @@ def CLF(X, y, verbose = False, show = False, save_fig = ''):
     X,y - features and labels
     '''
     
-    dct = {}
+    dic = {}
     clf_metrics = []
     LOG = ''
 
@@ -660,12 +659,72 @@ def CLF(X, y, verbose = False, show = False, save_fig = ''):
     rpt = ''    
     for v in zip(CLF_METRICS, clf_metrics):
         rpt += v[0] + "\t" + str(v[1]) + "\n"
-        dct[v[0]] = v[1]
+        dic[v[0]] = v[1]
         
     LOG += "\n\n" + rpt
         
-    return dct, IMG, LOG # acc_train, acc_test, acc_all # , vertice_set # vertice_set[0] is the decision boundary
+    return dic, IMG, LOG # acc_train, acc_test, acc_all # , vertice_set # vertice_set[0] is the decision boundary
+
+def SVM_Margin_Width(X, y, scale = False, show = False, save_fig=''):
+
+    '''
+    SVM hyperplane margin width
+
+    Note
+    ----
+    When the between-class distance is small (< 3std), there are many overlaps, 
+    the margin width is big. 
+    This metric is only linear after the distance is big enough.
+    '''
+
+    if scale:
+        X = MinMaxScaler().fit_transform(X)
+
+    svc_model = SVC(kernel='linear') # C = 10
+    svc_model = svc_model.fit(X, y)
+
+    support_vectors = svc_model.support_vectors_
+    w = svc_model.coef_[0]
+    p = np.linalg.norm(w, ord=2)
+    width = 2/p
+
+    IMG = ''
+
+    if len(support_vectors[1]) == 2 and len(set(y)) == 2:
+        df = pd.DataFrame(X)
+
+        x_min = np.min(df.iloc[:, 0]) - 0.5
+        x_max = np.max(df.iloc[:,  0]) + 0.5
+        y_min = np.min(df.iloc[:, 1]) - 0.5
+        y_max = np.max(df.iloc[:, 1]) + 0.5
+
+        x = np.arange(x_min, x_max, 0.1)
+        y0 = -svc_model.intercept_/w[1]-w[0]/w[1]*x
+        y_up = (1-svc_model.intercept_)/w[1]-w[0]/w[1]*x
+        y_down = (-1 - svc_model.intercept_) / w[1] - w[0] / w[1] * x
+
+        plt.plot(x, y0)
+        plt.plot(x, y_up, linestyle='--')
+        plt.plot(x, y_down, linestyle='--')
+        plt.ylim(y_min , y_max)
+        plt.xlim(x_min , x_max)
+
+        labels = set(y)
+
+        for label in labels:
+            cluster = X[np.where(y == label)]
+            plt.scatter(cluster[:, 0], cluster[:, 1])
+
+
+        IMG = plt2html(plt)
     
+        if show:
+            plt.show()
+        else:
+            plt.close()
+
+    return width, IMG
+
 def IG(X, y, show = False, save_fig = ''):
     """
     Return the feature-wise information gains.
@@ -731,7 +790,7 @@ def IG(X, y, show = False, save_fig = ''):
     
     return mi, IMG
 
-def CHISQ(X, y, verbose = False, show = False, save_fig = ''):
+def CHISQ(X, y, show = False, save_fig = ''):
     """
     Performa feature-wise chi-square test. 
     Returns an array of chi2 statistics and p-values on all the features.
@@ -781,6 +840,211 @@ def CHISQ(X, y, verbose = False, show = False, save_fig = ''):
 
     return ps.tolist(), CHI2s.tolist(), IMG
 
+
+def KW(X, y, verbose=False):
+    '''
+    A Kruskal-Wallis test is used to determine whether or not 
+    there is a statistically significant difference between 
+    the medians of three or more independent groups. 
+    This test is the nonparametric equivalent of the one-way ANOVA 
+    and is typically used when the normality assumption is violated.
+
+    Mann-Whitney或Wilcoxon检验比较两组，而Kruskal-Wallis检验比较3组及以上。
+    Kruskal-Wallis检验是一种非参数的单因素方差分析。它是基于秩（排序的，只考虑相对大小）的。
+
+    由于KW检验考虑了样本的排序信息,而不仅仅是大于或小于中位数,因此比median test具有更大的power
+    '''
+
+    if (len(set(y)) < 2):
+        print('The dataset must have at least 2 classes.')
+        return None, None
+
+    ps = []
+    Hs = []
+
+    for i in range(X.shape[1]):
+        Xi = X[:, i]
+        Xcis = []
+
+        for c in set(y):
+            Xc = Xi[y == c]
+            Xcis.append(Xc)
+
+        Xcis = np.array(Xcis)
+
+        if len(set(y)) == 2:
+            H, p = scipy.stats.kruskal(Xcis[0],Xcis[1],nan_policy='omit')
+        elif len(set(y)) == 3:
+            H, p = scipy.stats.kruskal(Xcis[0], Xcis[1], Xcis[2],nan_policy='omit')
+        elif len(set(y)) == 4:
+            H, p = scipy.stats.kruskal(Xcis[0], Xcis[1], Xcis[2],Xcis[3],nan_policy='omit')
+        elif len(set(y)) >= 5:
+            H, p = scipy.stats.kruskal(Xcis[0], Xcis[1], Xcis[2],Xcis[3],Xcis[4],nan_policy='omit')
+            print("WARN: only the first 5 classes will be analyzed.")
+
+        ps.append(p)
+        Hs.append(H)
+
+    if verbose:
+        print('The P values for X in dimensions 1 to {}:{}'.format(len(X[0]), ps))
+        print('The values of H statistics for X in dimensions 1 to {}:{}'.format(len(X[0]), Hs))
+
+    return ps,Hs
+
+def T_IND(X, y, verbose=False, show=False, max_plot_num = 5):
+    '''
+    independent t test. requires two classes/groups.
+    '''
+    
+    if (len(set(y)) != 2):
+        print('The dataset must have 2 classes.')
+        return None, None, None
+
+    ps=[]
+    Ts=[]
+    cnt = 0
+    IMG = ''
+
+    for i in range(X.shape[1]):
+        Xi = X[:, i]
+        Xcis = []
+
+        labels = []
+        for c in set(y):
+            Xc = Xi[y == c]
+            Xcis.append(Xc)
+            labels.append("$ X_" + str(i + 1) + "^{( y_" + str(c) + " )} $")
+
+        Xcis = np.array(Xcis)
+
+        bar=scipy.stats.bartlett(Xcis[0],Xcis[1])[1]
+        lev=scipy.stats.levene(Xcis[0],Xcis[1])[1]
+
+        if bar>0.5 or lev>0.5:
+            T, p = scipy.stats.ttest_ind(Xcis[0],Xcis[1])
+        else:
+            T, p = scipy.stats.ttest_ind(Xcis[0], Xcis[1],equal_var=False)
+
+        ps.append(p)
+        Ts.append(T)
+
+        if (cnt < max_plot_num):
+            plt.figure()
+            plt.boxplot(Xcis.T, notch=False, labels=labels) # plot ith feature of different classes   
+            test_result = "independent t test on X{}: T={},p={}".format(i + 1, round(T,3), round(p,3))
+            # plt.legend(labels)
+            plt.title(test_result)
+            IMG += plt2html(plt)
+    
+            if show:
+                plt.show()
+            else:
+                plt.close()
+        elif cnt == max_plot_num:
+            IMG += '<p>Showing the first ' + str(max_plot_num) + ' plots.</p>'     
+        else:
+            pass # plot no more to avoid memory cost
+
+    if verbose:
+        print('The P values of X in dimensions 1 to {}:{}'.format(len(X[0]), ps))
+        print('The values of T statistics of X in dimensions 1 to {}:{}'.format(len(X[0]), Ts))
+
+    return ps, Ts, IMG
+
+def MedianTest(X, y, verbose=False, show=False):
+    '''
+    中位数检验是独立性卡方检验的一种特殊情况。
+
+    给定k组样本，n1, n2 …… nk观测值，计算所有n1 +n2 + ……+nk观测值的中位数。
+    然后构造一个2xk列联表，其中第一行包含k个样本的中位数以上的观测值，第二行包含k个样本的中位数以下或等于中位数的观测值。
+    然后可以对该表应用独立性卡方检验。
+    '''
+    if (len(set(y)) < 2):
+        print('The dataset must have at least 2 classes.')
+        return None, None, None
+
+    ps = []
+    Ts = []
+    MED = [] # grand median
+    TBL = [] # contingency table
+    IMG = ''
+
+    for i in range(X.shape[1]):
+        Xi = X[:, i]
+        Xcis = []
+
+        for c in set(y):
+            Xc = Xi[y == c]
+            Xcis.append(Xc)
+
+        Xcis = np.array(Xcis)
+
+        if len(set(y)) == 2:
+            T, p, med, tbl = scipy.stats.median_test(Xcis[0], Xcis[1], ties='ignore')
+        elif len(set(y)) == 3:
+            T, p, med, tbl = scipy.stats.median_test(Xcis[0], Xcis[1], Xcis[2], ties='ignore')
+        elif len(set(y)) == 4:
+            T, p, med, tbl = scipy.stats.median_test(Xcis[0], Xcis[1], Xcis[2], Xcis[3], ties='ignore')
+        elif len(set(y)) >= 5:
+            T, p, med, tbl = scipy.stats.median_test(Xcis[0], Xcis[1], Xcis[2], Xcis[3], Xcis[4], ties='ignore')
+            print("WARN: only the first 5 classes will be analyzed.")
+
+        ps.append(p)
+        Ts.append(T)
+        MED.append(med)
+        TBL.append(tbl)
+
+        a1 = TBL[0][:, 0]
+        a2 = TBL[0][:, 1]
+
+    if len(X[0])==2 and len(set(y))==2:
+        
+        idx = 1
+
+        fig, axes = plt.subplots(1,2, figsize = (8,4))
+        
+        for ax, tbl, m, p, T in zip(axes, TBL, MED, ps, Ts):
+
+            tick_labels = ["> grand \n median", "< grand \n median"]
+            
+            im = ax.imshow(tbl, cmap='Greys')
+
+            # # We want to show all ticks...
+            ax.set_xticks(np.arange(len(tick_labels)))
+            ax.set_yticks(np.arange(len(tick_labels)))
+            # ... and label them with the respective list entries
+            ax.set_xticklabels(tick_labels)
+            ax.set_yticklabels(tick_labels)
+
+            # Rotate the tick labels and set their alignment.
+            plt.setp(ax.get_xticklabels(), rotation=0, ha="center", va="top",
+                    rotation_mode="anchor")
+            plt.setp(ax.get_yticklabels(), rotation=0, ha="right",
+                    rotation_mode="anchor")
+
+            # Loop over data dimensions and create text annotations.
+            for i in range(len(tick_labels)):
+                for j in range(len(tick_labels)):
+                    text = ax.text(j, i, tbl[i, j],
+                                ha="center", va="center", color="gray")
+
+            ax.set_title("Median contingency table (X" + str(idx) + ")\nCHI2 statistic = " + str(round(T,3)) + ", p = " + str(round(p,3)))
+            idx += 1
+        
+        fig.tight_layout()
+        IMG = plt2html(plt)
+    
+        if show:
+            plt.show()
+        else:
+            plt.close()
+            
+    if verbose:
+        print('The P value of X in dimensions 1 to {}:{}'.format(len(X[0]), ps))
+        print('The values of the contingency table chi2 statistic of X in dimensions 1 to {}:{}'.format(len(X[0]), Ts))
+
+    return Ts, ps, IMG
+
 def ANOVA(X,y, verbose = False, show = False, max_plot_num = 5):
     """
     Performa feature-wise ANOVA test. Returns an array of p-values on all the features and its minimum.
@@ -817,6 +1081,7 @@ def ANOVA(X,y, verbose = False, show = False, max_plot_num = 5):
             f,p= scipy.stats.f_oneway(Xcis[0], Xcis[1], Xcis[2], Xcis[3])
         elif (len(set(y)) >= 5): # if there are five or more classes
             f,p= scipy.stats.f_oneway(Xcis[0], Xcis[1], Xcis[2], Xcis[3], Xcis[4])
+            print('WARN: only the first 5 classes will be analyzed.')
         
         """
         Alternative implementation using sm.stats.anova_lm
@@ -1047,7 +1312,7 @@ def correlate(X,y, verbose = False, show = False):
     
     """       
 
-    dct = {}
+    dic = {}
     
     rs = []
     rhos = []
@@ -1083,26 +1348,26 @@ def correlate(X,y, verbose = False, show = False):
     if verbose:
         print(LOG)
 
-    dct['correlation.r'] = rs
-    dct['correlation.r2'] = np.power(rs,2) # R2, the R-squared effect size
-    dct['correlation.r.p'] = prs
-    dct['correlation.r.max'] = np.abs(rs).max() # abs max
-    dct['correlation.r.p.min'] = np.min(prs)
+    dic['correlation.r'] = rs
+    dic['correlation.r2'] = np.power(rs,2) # R2, the R-squared effect size
+    dic['correlation.r.p'] = prs
+    dic['correlation.r.max'] = np.abs(rs).max() # abs max
+    dic['correlation.r.p.min'] = np.min(prs)
     
-    dct['correlation.rho'] = rhos
-    dct['correlation.rho.p'] = prhos
-    dct['correlation.rho.max'] = np.abs(rhos).max() # abs max
-    dct['correlation.rho.p.min'] = np.min(prhos)
+    dic['correlation.rho'] = rhos
+    dic['correlation.rho.p'] = prhos
+    dic['correlation.rho.max'] = np.abs(rhos).max() # abs max
+    dic['correlation.rho.p.min'] = np.min(prhos)
     
-    dct['correlation.tau'] = taus
-    dct['correlation.tau.p'] = ptaus
-    dct['correlation.tau.max'] = np.abs(taus).max() # abs max
-    dct['correlation.tau.p.min'] = np.min(ptaus)
+    dic['correlation.tau'] = taus
+    dic['correlation.tau.p'] = ptaus
+    dic['correlation.tau.max'] = np.abs(taus).max() # abs max
+    dic['correlation.tau.p.min'] = np.min(ptaus)
 
     if (show):
 
         for key in ['correlation.r', 'correlation.r2', 'correlation.rho', 'correlation.tau']:
-            v = dct[key]
+            v = dic[key]
             if (X.shape[1] > 50):
                 plt.figure(figsize=(20,3))
             else:
@@ -1112,7 +1377,7 @@ def correlate(X,y, verbose = False, show = False):
             plt.title(key)
             plt.show()
     
-    return dct, LOG
+    return dic, LOG
 
 def KS(X,y, show = False, max_plot_num = 5):
     """
@@ -1271,12 +1536,12 @@ def ECoL_metrics(X,y):
             ''')
     
     rpt = ''
-    dct = {}
+    dic = {}
     for v in zip(ECoL_METRICS, metrics):
         rpt += v[0] + "\t" + str(v[1]) + "\n"
-        dct[v[0]] = v[1]
+        dic[v[0]] = v[1]
         
-    return dct, rpt
+    return dic, rpt
 
 def analyze_file(fn):
     if os.path.isfile(fn) == False:
@@ -1296,85 +1561,109 @@ def get_metrics(X,y):
     If we plot two PCs from PCA，the PCs will also be linearly uncorrelated, because they are the projections on two different orthogonal eigenvectors. 
     '''
     
-    dct,_,_ = CLF(X,y)
-    if dct is None:
-        dct = {}
+    dic,_,_ = CLF(X,y)
+    if dic is None:
+        dic = {}
  
     try:
         ber, _ = BER(X,y)
-        dct['classification.BER'] = ber
+        dic['classification.BER'] = ber
     except:
         print('Exception in GaussianNB.')
 
+    svm_width, _ = SVM_Margin_Width(X, y)
+    dic['classification.SVM.Margin'] = svm_width
+
     ig, _ = IG(X,y)
     if ig is not None:
-        dct['correlation.IG'] = ig
-        dct['correlation.IG.max'] = ig.max()
+        dic['correlation.IG'] = ig
+        dic['correlation.IG.max'] = ig.max()
 
-    dct_cor,_ = correlate(X,y)
-    dct.update(dct_cor)
+    dic_cor,_ = correlate(X,y)
+    dic.update(dic_cor)
 
     es, _ = cohen_d(X, y)
-    dct['test.ES'] = es
-    dct['test.ES.max'] = es.max()
+    dic['test.ES'] = es
+    dic['test.ES.max'] = es.max()
+
+    p, T, _ = T_IND(X,y)
+    dic['test.student'] = p
+    dic['test.student.min'] = np.min(p)
+    dic['test.student.min.log10'] = np.log10 (np.min(p) )
+    dic['test.student.T'] = T
+    dic['test.student.T.max'] = np.max(T)
 
     p, F, _ = ANOVA(X,y)
-    dct['test.ANOVA'] = p
-    dct['test.ANOVA.min'] = np.min(p)
-    dct['test.ANOVA.min.log10'] = np.log10 (np.min(p) )
-    dct['test.ANOVA.F'] = F
-    dct['test.ANOVA.F.max'] = np.max(F)
+    dic['test.ANOVA'] = p
+    dic['test.ANOVA.min'] = np.min(p)
+    dic['test.ANOVA.min.log10'] = np.log10 (np.min(p) )
+    dic['test.ANOVA.F'] = F
+    dic['test.ANOVA.F.max'] = np.max(F)
 
     p, F, log = MANOVA(X,y)
     if log == 'Exception in MANOVA':
         pass
     else:
-        dct['test.MANOVA'] = p
-        dct['test.MANOVA.log10'] = np.log10 (p)
-        dct['test.MANOVA.F'] = F
+        dic['test.MANOVA'] = p
+        dic['test.MANOVA.log10'] = np.log10 (p)
+        dic['test.MANOVA.F'] = F
 
     p, U, _ = MWW(X,y)
-    dct['test.MWW'] = p
-    dct['test.MWW.min'] = np.min(p)
-    dct['test.MWW.min.log10'] = np.log10 (np.min(p) )
-    dct['test.MWW.U'] = U
-    dct['test.MWW.U.min'] = np.min(U) 
+    dic['test.MWW'] = p
+    dic['test.MWW.min'] = np.min(p)
+    dic['test.MWW.min.log10'] = np.log10 (np.min(p) )
+    dic['test.MWW.U'] = U
+    dic['test.MWW.U.min'] = np.min(U) 
 
     p, D, _ = KS(X,y)
-    dct['test.KS'] = p
-    dct['test.KS.min'] = np.min(p)
-    dct['test.KS.min.log10'] = np.log10 (np.min(p) )
-    dct['test.KS.D'] = D
-    dct['test.KS.D.max'] = np.max(D) 
+    dic['test.KS'] = p
+    dic['test.KS.min'] = np.min(p)
+    dic['test.KS.min.log10'] = np.log10 (np.min(p) )
+    dic['test.KS.D'] = D
+    dic['test.KS.D.max'] = np.max(D) 
     
     p, C, _ = CHISQ(X,y)
-    dct['test.CHISQ'] = p
-    dct['test.CHISQ.min'] = np.min(p)
-    dct['test.CHISQ.min.log10'] = np.log10 (np.min(p))
-    dct['test.CHISQ.CHI2'] = C
-    dct['test.CHISQ.CHI2.max'] = np.max(C)
+    dic['test.CHISQ'] = p
+    dic['test.CHISQ.min'] = np.min(p)
+    dic['test.CHISQ.min.log10'] = np.log10 (np.min(p))
+    dic['test.CHISQ.CHI2'] = C
+    dic['test.CHISQ.CHI2.max'] = np.max(C)
+
+    p, H = KW(X,y)
+    dic['test.KW'] = p
+    dic['test.KW.min'] = np.min(p)
+    dic['test.KW.min.log10'] = np.log10 (np.min(p))
+    dic['test.KW.H'] = H
+    dic['test.KW.H.max'] = np.max(H)
+
+    p, T, _ = MedianTest(X,y)
+    dic['test.Median'] = p
+    dic['test.Median.min'] = np.min(p)
+    dic['test.Median.min.log10'] = np.log10 (np.min(p))
+    dic['test.Median.CH2'] = T
+    dic['test.Median.CH2.max'] = np.max(T)
 
     if ENABLE_R:
         try:
-            dct_ecol,_ = ECoL_metrics(X,y)
-            dct.update(dct_ecol)
+            dic_ecol,_ = ECoL_metrics(X,y)
+            dic.update(dic_ecol)
         except Exception as e:
             print(e)
 
-    dct_s = {}
+    dic_s = {}
     
-    for k, v in dct.items():
+    for k, v in dic.items():
         if hasattr(v, "__len__"): # this is an np array or list
-            dct[k] = list(v)
+            dic[k] = list(v)
         else: # this only contains single-value metrics
-            dct_s[k] = v
+            dic_s[k] = v
 
-    return dct, dct_s
+    return dic, dic_s
 
 def metrics_keys():
     X, y = mvg(md = 2, nobs = 10)
-    dct = get_metrics(X,y)
-    return list(dct[0].keys())
+    dic = get_metrics(X,y)
+    return list(dic[0].keys())
 
 def get_json(X,y):    
     return json.dumps(get_metrics(X,y))
@@ -1397,6 +1686,11 @@ def get_html(X,y):
         html += tr
     except:
         print('Exception in GaussianNB.')
+
+    svm_margin, svm_margin_img = SVM_Margin_Width(X,y,show = False)
+
+    tr = '<tr><td> SVM Margin Width = ' + str(svm_margin) + '<br/>' + svm_margin_img + '</td><tr>'
+    html += tr
     
     clf, clf_img, clf_log = CLF(X,y,show = False)
 
@@ -1413,6 +1707,11 @@ def get_html(X,y):
     tr = '<tr><td><pre>' + corr_log + '</pre></td><tr>'
     html += tr
 
+    t_p, _, t_img = T_IND(X,y)
+    
+    tr = '<tr><td> Independent t-test p' + str(t_p) + '<br/>' + t_img + '</td><tr>'
+    html += tr     
+    
     anova_p, _, anova_img = ANOVA(X,y)
     
     tr = '<tr><td> ANOVA p' + str(anova_p) + '<br/>' + anova_img + '</td><tr>'
@@ -1439,9 +1738,18 @@ def get_html(X,y):
     chi2s_p,_, chi2s_img = CHISQ(X,y)
     
     tr = '<tr><td> CHISQ p = ' + str(chi2s_p) + '<br/>' + chi2s_img + '</td><tr>'
+    html += tr   
+
+    m_p,_, m_img = MedianTest(X,y)
+    
+    tr = '<tr><td> Median test p = ' + str(m_p) + '<br/>' + m_img + '</td><tr>'
+    html += tr 
+
+    kw_p,_ = KW(X,y)
+    
+    tr = '<tr><td> Kruskal-Wallis test p = ' + str(kw_p) + '</td><tr>'
     html += tr    
-    
-    
+        
     es, es_img = cohen_d(X,y)
 
     tr = '<tr><td> ES = ' + str(es) + '<br/>' + es_img + '</td><tr>'
@@ -1469,11 +1777,16 @@ def get_html(X,y):
     return html
 
 
-# try different sample sizes (nobs)
 def simulate(mds, repeat = 1, nobs = 100, dims = 2):
+    '''
+    Try different mds (between-group distances)
 
-    dcts = {}
-    
+    Parameters
+    ----------
+    mds : an array. between-classes mean distances    
+    '''
+
+    dic = {}    
 
     ## splits (divide 1 std into how many sections) * repeat         
     pbar = tqdm(total = len(mds) * repeat, position=0)
@@ -1483,7 +1796,7 @@ def simulate(mds, repeat = 1, nobs = 100, dims = 2):
 
     for md in mds:
         
-        dct = {}
+        raw_dic = {}
         
         for i in range(repeat):
             X,y = mvg(
@@ -1497,111 +1810,126 @@ def simulate(mds, repeat = 1, nobs = 100, dims = 2):
             ## if detailed:
             #    print('d = ', round(md,3))
             
-            _, dct1 = get_metrics(X,y)
-            for k, v in dct1.items():
-                if k in dct:
-                    dct[k].append(v) # dct[k] = dct[k] + v
+            _, dic1 = get_metrics(X,y)
+            for k, v in dic1.items():
+                if k in raw_dic:
+                    raw_dic[k].append(v) # dic[k] = dic[k] + v
                 else:
-                    dct[k] = [v] # v
+                    raw_dic[k] = [v] # v
             
             pbar.update()
             ## End of inner iteration ##
 
-        for k, v in dct.items():
+        for k, v in raw_dic.items():
 
             trim_size = int(repeat / 10) 
 
             if (repeat > 10): # remove the max and min
-                dct[k] = np.mean(sorted(v)[trim_size:-trim_size])
+                raw_dic[k] = np.mean(sorted(v)[trim_size:-trim_size])
             else:
-                dct[k] = np.mean(v)
+                raw_dic[k] = np.mean(v)
             
         ## End of outer iteration ##
     
-        for k, v in dct.items():
-            if k in dcts:
-                dcts[k] = np.append(dcts[k], v)
+        for k, v in raw_dic.items():
+            if k in dic:
+                dic[k] = np.append(dic[k], v)
             else:
-                dcts[k] = np.array([v])
+                dic[k] = np.array([v])
     
-    dcts['d'] = np.array(mds)
+    dic['d'] = np.array(mds)
     
-    return dcts
+    return dic
 
-def visualize_dcts(dcts):
+def generate_html_for_dict(dic):
+    s = '<table>'
+    for k in dic:
+        s += '<tr>'
+        s += '<td>' + str(k) + '</td>'
+        for item in dic[k]:            
+            s+= '<td>' + str(round(item,3)) + '</td>' # item[0] is key, item[1] is value           
+        s+= '</tr>'
+    s += '</table>'
+    return s
+
+def visualize_dict(dic):
+    '''
+    Visualize the metric dictionary returned from simulate()  
+    '''
     
-    N = len(dcts) - 1
+    N = len(dic) - 1
     
     fig = plt.figure(figsize=(48, 10*(N/10+1)))
     plt.rcParams.update({'font.size': 18})
     
     i = 0
-    for k, v in dcts.items():
+    for k, v in dic.items():
         if k == 'd':
             pass
         else:
             ax = fig.add_subplot(round(N/6+1), 6, i+1)
-            ax.scatter(dcts['d'], v, label = k)
-            ax.plot(dcts['d'], v)
+            ax.scatter(dic['d'], v, label = k)
+            ax.plot(dic['d'], v)
             ax.xaxis.set_major_locator(mticker.MultipleLocator(1))
             ax.legend() # loc='upper center'
             ax.axis('tight')
-        i += 1
+            i += 1
         
     plt.axis('tight')
     plt.show()
 
     plt.rcParams.update({'font.size': 10})
     
-def visualize_corr_matrix(dcts, cmap = 'coolwarm', exclude_d = True):
+def visualize_corr_matrix(dic, cmap = 'coolwarm', threshold = 0.25):
     
-    M = dcts['d']
+    M = dic['d']
     names = ['d']
     
-    for k, v in dcts.items():
-        if k == 'd' or len(v) != len(dcts['d']):
+    for k, v in dic.items():
+        if k == 'd' or len(v) != len(dic['d']):
             pass
         else:
             M = np.vstack((M, v))
             names.append(k)
-            
-    # exclude the d column
-    if (exclude_d):
-        M = M[1:]
-        names = names[1:]
 
     plt.figure(figsize = (36,30))
     dfM = pd.DataFrame(M.T, columns = names) #, columns = ['d', 'BER', 'ACC', 'TOR', 'IG', 'MAN'])
 
     # plot the heatmap
     # sns.heatmap(np.abs(dfM.corr()), cmap="YlGnBu", annot=True)
-    sns.heatmap(dfM.corr(), cmap=cmap, annot=True) # use diverging colormap, e.g., seismic, coolwarm, bwr
+    sns.heatmap(dfM.corr(), cmap=cmap, annot=True) # should use diverging colormap, e.g., seismic, coolwarm, bwr
 
     plt.tick_params(axis='both', which='major', labelsize=10)
     plt.tick_params(axis='both', which='minor', labelsize=10)
     plt.rcParams['xtick.labelsize'] = 10
     plt.rcParams['ytick.labelsize'] = 10
     # plt.rcParams.update({'font.size': 10})   # restore fontsize
+    ## https://stackoverflow.com/questions/56942670/matplotlib-seaborn-first-and-last-row-cut-in-half-of-heatmap-plot
+    # bottom, top = ax.get_ylim()
+    # ax.set_ylim(bottom + 0.5, top - 0.5)
     plt.show()
     
     plt.figure(figsize = (20, 7))
-    plt.bar(names[1:], np.abs(dfM.corr().values[0,1:]), facecolor="none", edgecolor = "black", width = 0.8, label = 'correlation coefficient (abs)') # ,width = 0.6, hatch='x'
-    plt.plot(names[1:], [0.9] * len(names[1:]), '--', label = 'threshold at 0.9')
-    plt.title('correlation with between-class distance')
+    plt.bar(names[1:], np.square(dfM.corr().values[0,1:]), 
+    facecolor="none", edgecolor = "black", 
+    width = 0.8, label = 'R2 effect size') # ,width = 0.6, hatch='x'
+    plt.plot(names[1:], [threshold] * len(names[1:]), '--', 
+    label = 'R2 threshold = ' + str(threshold))
+    plt.title('R2 effect size of the correlation with between-class distance')
     plt.xticks(rotation = 90)
     plt.legend(loc = 'lower right')
     plt.show()
-    
+
     # print(np.where( np.abs(dfM.corr().values[0,1:])>0.9 ))
-    print('Metrics above the threshold: ', np.array(names[1:]) [np.where( np.abs(dfM.corr().values[0,1:])>0.9 )])
+    print('Metrics above the threshold (' + str(threshold) + '): ', np.array(names[1:]) [np.where( np.square(dfM.corr().values[0,1:])>threshold )])
 
-def extract_PC(dcts):
+def extract_PC(dic):
 
-    M = dcts['d']
+    M = dic['d']
     names = []
     
-    for k, v in dcts.items():
-        if k == 'd' or len(v) != len(dcts['d']) or np.isnan(v).any():
+    for k, v in dic.items():
+        if k == 'd' or len(v) != len(dic['d']) or np.isnan(v).any():
             pass
         else:
             M = np.vstack((M, v))
